@@ -138,24 +138,6 @@ func shortcircuitBlock(b *Block) bool {
 	if len(b.Values) != nval+nOtherPhi {
 		return false
 	}
-	if nOtherPhi > 0 {
-		// Check for any phi which is the argument of another phi.
-		// These cases are tricky, as substitutions done by replaceUses
-		// are no longer trivial to do in any ordering. See issue 45175.
-		m := make(map[*Value]bool, 1+nOtherPhi)
-		for _, v := range b.Values {
-			if v.Op == OpPhi {
-				m[v] = true
-			}
-		}
-		for v := range m {
-			for _, a := range v.Args {
-				if a != v && m[a] {
-					return false
-				}
-			}
-		}
-	}
 
 	// Locate index of first const phi arg.
 	cidx := -1
@@ -279,11 +261,6 @@ func shortcircuitBlock(b *Block) bool {
 // and the CFG modifications must not proceed.
 // The returned function assumes that shortcircuitBlock has completed its CFG modifications.
 func shortcircuitPhiPlan(b *Block, ctl *Value, cidx int, ti int64) func(*Value, int) {
-	const go115shortcircuitPhis = true
-	if !go115shortcircuitPhis {
-		return nil
-	}
-
 	// t is the "taken" branch: the successor we always go to when coming in from p.
 	t := b.Succs[ti].b
 	// u is the "untaken" branch: the successor we never go to when coming in from p.
